# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
# SPDX-FileCopyrightText: 2023-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Diego Badillo <diego.badillo@sansano.usm.cl>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for the Dotbot protocol API."""

import dataclasses
from abc import ABC, abstractmethod
from binascii import hexlify
from dataclasses import dataclass
from enum import Enum, IntEnum
from itertools import chain
from typing import List

PROTOCOL_VERSION = 9


class PayloadType(Enum):
    """Types of DotBot payload types."""

    CMD_MOVE_RAW = 0
    CMD_RGB_LED = 1
    LH2_RAW_DATA = 2
    LH2_LOCATION = 3
    ADVERTISEMENT = 4
    GPS_POSITION = 5
    DOTBOT_DATA = 6
    CONTROL_MODE = 7
    LH2_WAYPOINTS = 8
    GPS_WAYPOINTS = 9
    SAILBOT_DATA = 10
    CMD_XGO_ACTION = 11
    LH2_PROCESSED_DATA = 12
    INVALID_PAYLOAD = 13  # Increase each time a new payload type is added
    DOTBOT_SIMULATOR_DATA = 250


class ApplicationType(IntEnum):
    """Types of DotBot applications."""

    DotBot = 0  # pylint: disable=invalid-name
    SailBot = 1  # pylint: disable=invalid-name
    Freebot = 2  # pylint: disable=invalid-name
    XGO = 3
    LH2_mini_mote = 4


class ControlModeType(IntEnum):
    """Types of DotBot control modes."""

    MANUAL = 0
    AUTO = 1


class ProtocolPayloadParserException(Exception):
    """Exception raised on invalid or unsupported payload."""


@dataclass
class ProtocolField:
    """Data class that describes a payload field."""

    value: int = 0
    name: str = ""
    length: int = 1
    signed: bool = False


@dataclass
class ProtocolData(ABC):
    """Base class for protocol payload data classes."""

    @property
    @abstractmethod
    def fields(self) -> List[ProtocolField]:
        """Returns the list of fields in this data."""

    @staticmethod
    @abstractmethod
    def from_bytes(bytes_):
        """Returns a ProtocolData instance from a bytearray."""


@dataclass
class ProtocolHeader(ProtocolData):
    """Dataclass that holds header fields."""

    destination: int = 0xFFFFFFFFFFFFFFFF
    source: int = 0x0000000000000000
    swarm_id: int = 0x0000
    application: ApplicationType = ApplicationType.DotBot
    version: int = PROTOCOL_VERSION
    msg_id: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.destination, name="dst", length=8),
            ProtocolField(self.source, name="src", length=8),
            ProtocolField(self.swarm_id, name="swarm id", length=2),
            ProtocolField(self.application, name="app."),
            ProtocolField(self.version, name="ver."),
            ProtocolField(self.msg_id, name="msg id", length=4),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return ProtocolHeader(
            int.from_bytes(bytes_[0:8], "little"),  # destination
            int.from_bytes(bytes_[8:16], "little"),  # source
            int.from_bytes(bytes_[16:18], "little"),  # swarm_id
            ApplicationType(int.from_bytes(bytes_[18:19], "little")),  # application
            int.from_bytes(bytes_[19:20], "little"),  # version
            int.from_bytes(bytes_[20:24], "little"),  # msg id
        )


@dataclass
class CommandMoveRaw(ProtocolData):
    """Dataclass that holds move raw command data fields."""

    left_x: int = 0
    left_y: int = 0
    right_x: int = 0
    right_y: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.left_x, name="lx", length=1, signed=True),
            ProtocolField(self.left_y, name="ly", length=1, signed=True),
            ProtocolField(self.right_x, name="rx", length=1, signed=True),
            ProtocolField(self.right_y, name="ry", length=1, signed=True),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return CommandMoveRaw(*bytes_[0:4])


@dataclass
class CommandRgbLed(ProtocolData):
    """Dataclass that holds a complete rgb led command fields."""

    red: int = 0
    green: int = 0
    blue: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.red, name="red"),
            ProtocolField(self.green, name="green"),
            ProtocolField(self.blue, name="blue"),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return CommandRgbLed(*bytes_[0:3])


@dataclass
class CommandXgoAction(ProtocolData):
    """Dataclass that holds an XGO action."""

    action: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.action, name="action"),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return CommandXgoAction(bytes_[0])


@dataclass
class Lh2RawLocation(ProtocolData):
    """Dataclass that holds LH2 raw location data."""

    bits: int = 0x0000000000000000
    polynomial_index: int = 0x00
    offset: int = 0x00

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.bits, name="bits", length=8),
            ProtocolField(self.polynomial_index, name="poly", length=1),
            ProtocolField(self.offset, name="off.", length=1, signed=True),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return Lh2RawLocation(
            int.from_bytes(bytes_[0:8], "little"),
            int.from_bytes(bytes_[8:9], "little"),
            int.from_bytes(bytes_[9:10], "little", signed=True),
        )


@dataclass
class Lh2ProcessedLocation(ProtocolData):
    """Dataclass that holds LH2 processed location data."""

    polynomial_index: int = 0x00
    lfsr_index: int = 0x00000000
    timestamp_us: int = 0x00000000

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.polynomial_index, name="poly", length=1),
            ProtocolField(self.lfsr_index, name="lfsr_index", length=4),
            ProtocolField(self.timestamp_us, name="timestamp_us", length=4),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return Lh2ProcessedLocation(
            int.from_bytes(bytes_[0:1], "little"),
            int.from_bytes(bytes_[1:5], "little"),
            int.from_bytes(bytes_[5:9], "little"),
        )


@dataclass
class Lh2RawData(ProtocolData):
    """Dataclass that holds LH2 raw data."""

    locations: List[Lh2RawLocation] = dataclasses.field(default_factory=lambda: [])

    @property
    def fields(self) -> List[ProtocolField]:
        return list(chain(*[location.fields for location in self.locations]))

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return Lh2RawData(
            [
                Lh2RawLocation.from_bytes(bytes_[0:10]),
                Lh2RawLocation.from_bytes(bytes_[10:20]),
            ]
        )


@dataclass
class LH2Location(ProtocolData):
    """Dataclass that holds LH2 computed location data."""

    pos_x: int = 0
    pos_y: int = 0
    pos_z: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.pos_x, name="x", length=4),
            ProtocolField(self.pos_y, name="y", length=4),
            ProtocolField(self.pos_z, name="z", length=4),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return LH2Location(
            int.from_bytes(bytes_[0:4], "little"),
            int.from_bytes(bytes_[4:8], "little"),
            int.from_bytes(bytes_[8:12], "little"),
        )


@dataclass
class DotBotData(ProtocolData):
    """Dataclass that holds direction and LH2 raw data from DotBot application."""

    direction: int = 0xFFFF
    locations: List[Lh2RawLocation] = dataclasses.field(default_factory=lambda: [])

    @property
    def fields(self) -> List[ProtocolField]:
        _fields = [ProtocolField(self.direction, name="dir.", length=2, signed=True)]
        _fields += list(chain(*[location.fields for location in self.locations]))
        return _fields

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return DotBotData(
            direction=int.from_bytes(bytes_[0:2], "little", signed=True),
            locations=[
                Lh2RawLocation.from_bytes(bytes_[2:12]),
                Lh2RawLocation.from_bytes(bytes_[12:22]),
            ],
        )


@dataclass
class GPSPosition(ProtocolData):
    """Dataclass that holds GPS positions."""

    latitude: int = 0
    longitude: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.latitude, name="latitude", length=4, signed=True),
            ProtocolField(self.longitude, name="longitude", length=4, signed=True),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return GPSPosition(
            latitude=int.from_bytes(bytes_[0:4], "little", signed=True),
            longitude=int.from_bytes(bytes_[4:8], "little", signed=True),
        )


@dataclass
class SailBotData(ProtocolData):
    """Dataclass that holds SailBot data from SailBot application."""

    direction: int = 0xFFFF
    latitude: int = 0
    longitude: int = 0
    wind_angle: int = 0xFFFF  # uint angles from 0 to 359
    rudder_angle: int = 0
    sail_angle: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.direction, name="dir.", length=2, signed=False),
            ProtocolField(self.latitude, name="latitude", length=4, signed=True),
            ProtocolField(self.longitude, name="longitude", length=4, signed=True),
            ProtocolField(self.wind_angle, name="wind ang", length=2, signed=False),
            ProtocolField(self.rudder_angle, name="rud.", length=1, signed=True),
            ProtocolField(self.sail_angle, name="sail.", length=1, signed=True),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return SailBotData(
            direction=int.from_bytes(bytes_[0:2], "little", signed=False),
            latitude=int.from_bytes(bytes_[2:6], "little", signed=True),
            longitude=int.from_bytes(bytes_[6:10], "little", signed=True),
            wind_angle=int.from_bytes(bytes_[10:12], "little", signed=False),
            rudder_angle=int.from_bytes(bytes_[12:13], "little", signed=True),
            sail_angle=int.from_bytes(bytes_[13:14], "little", signed=True),
        )


@dataclass
class DotBotSimulatorData(ProtocolData):
    """Dataclass that holds direction and GPS data and heading from SailBot application."""

    theta: int = 0xFFFF
    pos_x: int = 0
    pos_y: int = 0

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.theta, name="theta", length=2),
            ProtocolField(self.pos_x, name="pos_x", length=4),
            ProtocolField(self.pos_y, name="pos_y", length=4),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return DotBotSimulatorData(
            theta=int.from_bytes(bytes_[0:2], "little"),
            pos_x=int.from_bytes(bytes_[2:6], "little"),
            pos_y=int.from_bytes(bytes_[6:10], "little"),
        )


@dataclass
class Advertisement(ProtocolData):
    """Dataclass that holds an advertisement (emtpy)."""

    @property
    def fields(self) -> List[ProtocolField]:
        return []

    @staticmethod
    def from_bytes(_: bytes) -> ProtocolData:
        return Advertisement()


@dataclass
class ControlMode(ProtocolData):
    """Dataclass that holds a control mode message."""

    mode: ControlModeType = ControlModeType.MANUAL

    @property
    def fields(self) -> List[ProtocolField]:
        return [
            ProtocolField(self.mode, "mode"),
        ]

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        return ControlMode(bytes_[0])


@dataclass
class LH2Waypoints(ProtocolData):
    """Dataclass that holds a list of LH2 waypoints."""

    threshold: int
    waypoints: List[LH2Location] = dataclasses.field(default_factory=lambda: [])

    @property
    def fields(self) -> List[ProtocolField]:
        _fields = [ProtocolField(len(self.waypoints), name="len.")]
        _fields += [ProtocolField(value=self.threshold, name="thr.")]
        _fields += list(chain(*[waypoint.fields for waypoint in self.waypoints]))
        return _fields

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        waypoints_count = int(bytes_[0])
        threshold = int(bytes_[1])
        waypoints_bytes = bytes_[2:]
        waypoints = []
        for idx in range(waypoints_count):
            for i in range(3):
                waypoints.append(
                    int.from_bytes(
                        waypoints_bytes[12 * idx + i * 4 : 12 * idx + (i + 1) * 4],
                        byteorder="little",
                    )
                )
        waypoints = [
            (waypoints[i], waypoints[i + 1], waypoints[i + 2])
            for i in range(0, 3 * waypoints_count, 3)
        ]
        return LH2Waypoints(threshold=threshold, waypoints=waypoints)


@dataclass
class GPSWaypoints(ProtocolData):
    """Dataclass that holds a list of GPS waypoints."""

    threshold: int
    waypoints: List[GPSPosition] = dataclasses.field(default_factory=lambda: [])

    @property
    def fields(self) -> List[ProtocolField]:
        _fields = [ProtocolField(len(self.waypoints), name="len.")]
        _fields += [ProtocolField(value=self.threshold, name="thr.")]
        _fields += list(chain(*[waypoint.fields for waypoint in self.waypoints]))
        return _fields

    @staticmethod
    def from_bytes(bytes_) -> ProtocolData:
        waypoints_count = int(bytes_[0])
        threshold = int(bytes_[1])
        waypoints_bytes = bytes_[2:]
        waypoints = []
        for idx in range(2 * waypoints_count):
            waypoints.append(
                float(
                    int.from_bytes(
                        waypoints_bytes[4 * idx : 4 * (idx + 1)], byteorder="little"
                    )
                    / 1e6
                )
            )
        waypoints = [
            (waypoints[i], waypoints[i + 1]) for i in range(0, 2 * waypoints_count, 2)
        ]
        return GPSWaypoints(threshold=threshold, waypoints=waypoints)


@dataclass
class ProtocolPayload:
    """Manage a protocol complete payload (header + type + values)."""

    header: ProtocolHeader
    payload_type: PayloadType
    values: ProtocolData

    def to_bytes(self, endian="little") -> bytes:
        """Converts a payload to a bytearray."""
        buffer = bytearray()
        for field in self.header.fields:
            buffer += int(field.value).to_bytes(
                length=field.length, byteorder=endian, signed=field.signed
            )
        buffer += int(self.payload_type.value).to_bytes(length=1, byteorder=endian)
        for field in self.values.fields:
            buffer += int(field.value).to_bytes(
                length=field.length, byteorder=endian, signed=field.signed
            )
        return buffer

    @staticmethod
    def from_bytes(bytes_: bytes):
        """Parse a bytearray to return a protocol payload instance."""
        try:
            header = ProtocolHeader.from_bytes(bytes_[0:24])
        except ValueError as exc:
            raise ProtocolPayloadParserException(f"Invalid header: {exc}") from exc
        if header.version != PROTOCOL_VERSION:
            raise ProtocolPayloadParserException(
                f"Invalid header: Unsupported payload version '{header.version}' (expected: {PROTOCOL_VERSION})"
            )
        payload_type = PayloadType(int.from_bytes(bytes_[24:25], "little"))
        if payload_type == PayloadType.CMD_MOVE_RAW:
            values = CommandMoveRaw.from_bytes(bytes_[25:30])
        elif payload_type == PayloadType.CMD_RGB_LED:
            values = CommandRgbLed.from_bytes(bytes_[25:29])
        elif payload_type == PayloadType.LH2_RAW_DATA:
            values = Lh2RawData.from_bytes(bytes_[25:45])
        elif payload_type == PayloadType.LH2_LOCATION:
            values = LH2Location.from_bytes(bytes_[25:37])
        elif payload_type == PayloadType.ADVERTISEMENT:
            values = Advertisement.from_bytes(None)
        elif payload_type == PayloadType.GPS_POSITION:
            values = GPSPosition.from_bytes(bytes_[25:33])
        elif payload_type == PayloadType.DOTBOT_DATA:
            values = DotBotData.from_bytes(bytes_[25:47])
        elif payload_type == PayloadType.SAILBOT_DATA:
            values = SailBotData.from_bytes(bytes_[25:39])
        elif payload_type == PayloadType.DOTBOT_SIMULATOR_DATA:
            values = DotBotSimulatorData.from_bytes(bytes_[25:35])
        elif payload_type == PayloadType.CONTROL_MODE:
            values = ControlMode.from_bytes(bytes_[25:26])
        elif payload_type == PayloadType.LH2_WAYPOINTS:
            values = LH2Waypoints.from_bytes(bytes_[25:])
        elif payload_type == PayloadType.GPS_WAYPOINTS:
            values = GPSWaypoints.from_bytes(bytes_[25:])
        elif payload_type == PayloadType.LH2_PROCESSED_DATA:
            values = Lh2ProcessedLocation.from_bytes(bytes_[25:34])
        else:
            raise ProtocolPayloadParserException(
                f"Unsupported payload type '{payload_type.value}'"
            )
        return ProtocolPayload(header, payload_type, values)

    def __repr__(self):
        header_separators = [
            "-" * (4 * field.length + 2) for field in self.header.fields
        ]
        type_separators = ["-" * 6]  # type
        values_separators = [
            "-" * (4 * field.length + 2) for field in self.values.fields
        ]
        header_names = [
            f" {field.name:<{4 * field.length + 1}}" for field in self.header.fields
        ]
        type_name = [" type "]
        values_names = [
            f" {field.name:<{4 * field.length + 1}}" for field in self.values.fields
        ]
        header_values = [
            f" 0x{hexlify(int(field.value).to_bytes(field.length, 'big', signed=field.signed)).decode():<{4 * field.length - 1}}"
            for field in self.header.fields
        ]
        type_value = [
            f" 0x{hexlify(int(PayloadType(self.payload_type).value).to_bytes(1, 'big')).decode():<3}"
        ]
        values_values = [
            f" 0x{hexlify(int(field.value).to_bytes(field.length, 'big', signed=field.signed)).decode():<{4 * field.length - 1}}"
            for field in self.values.fields
        ]
        num_bytes = (
            sum(field.length for field in self.header.fields)
            + 1
            + sum(field.length for field in self.values.fields)
        )
        if num_bytes > 32:
            # put values on a separate row
            separators = header_separators + type_separators
            names = header_names + type_name
            values = header_values + type_value
            return (
                f" {' ' * 16}+{'+'.join(separators)}+\n"
                f" {PayloadType(self.payload_type).name:<16}|{'|'.join(names)}|\n"
                f" {f'({num_bytes} Bytes)':<16}|{'|'.join(values)}|\n"
                f" {' ' * 16}+{'+'.join(separators)}+\n"
                f" {' ' * 16}+{'+'.join(values_separators)}+\n"
                f" {' ' * 16}|{'|'.join(values_names)}|\n"
                f" {' ' * 16}|{'|'.join(values_values)}|\n"
                f" {' ' * 16}+{'+'.join(values_separators)}+\n"
            )

        # all in a row by default
        separators = header_separators + type_separators + values_separators
        names = header_names + type_name + values_names
        values = header_values + type_value + values_values
        return (
            f" {' ' * 16}+{'+'.join(separators)}+\n"
            f" {PayloadType(self.payload_type).name:<16}|{'|'.join(names)}|\n"
            f" {f'({num_bytes} Bytes)':<16}|{'|'.join(values)}|\n"
            f" {' ' * 16}+{'+'.join(separators)}+\n"
        )



# SPDX-FileCopyrightText: 2023-present Inria
# SPDX-FileCopyrightText: 2023-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Dotbot simulator for the DotBot project."""

import threading
import time
from dataclasses import dataclass
from enum import Enum
from math import atan2, cos, pi, sin, sqrt
from typing import Callable

from dotbot import GATEWAY_ADDRESS_DEFAULT, SWARM_ID_DEFAULT
from dotbot.hdlc import hdlc_decode, hdlc_encode
from dotbot.logger import LOGGER
from dotbot.protocol import (
    PROTOCOL_VERSION,
    Advertisement,
    ApplicationType,
    DotBotSimulatorData,
    PayloadType,
    ProtocolHeader,
    ProtocolPayload,
)

R = 1
L = 2
SIMULATOR_STEP_DELTA_T = 0.005


def diff_drive_bot(x_pos_old, y_pos_old, theta_old, v_right, v_left):
    """Execute state space model of a rigid differential drive robot."""
    x_dot = R / 2 * (v_right + v_left) * cos(theta_old - pi) * 50000
    y_dot = R / 2 * (v_right + v_left) * sin(theta_old - pi) * 50000
    theta_dot = R / L * (-v_right + v_left)

    x_pos = x_pos_old + x_dot * SIMULATOR_STEP_DELTA_T
    y_pos = y_pos_old + y_dot * SIMULATOR_STEP_DELTA_T
    theta = (theta_old + theta_dot * SIMULATOR_STEP_DELTA_T) % (2 * pi)

    return x_pos, y_pos, theta


class DotBotSimulatorMode(Enum):
    """Operation mode of the dotbot simulator."""

    MANUAL = "MANUAL"
    AUTOMATIC = "AUTOMATIC"


@dataclass
class Waypoint:
    """Waypoint class for the dotbot simulator."""

    x: int
    y: int


class DotBotSimulator:
    """Simulator class for the dotbot."""

    def __init__(self, address):
        self.address = address
        self.pos_x = 0.5 * 1e6
        self.pos_y = 0.5 * 1e6
        self.theta = 0

        self.v_left = 0
        self.v_right = 0

        self.waypoint_threshold = 0
        self.waypoints = []
        self.waypoint_index = 0

        self.controller_mode: DotBotSimulatorMode = DotBotSimulatorMode.MANUAL
        self.logger = LOGGER.bind(context=__name__, address=self.address)

    @property
    def header(self):
        return ProtocolHeader(
            destination=int(GATEWAY_ADDRESS_DEFAULT, 16),
            source=int(self.address, 16),
            swarm_id=int(SWARM_ID_DEFAULT, 16),
            application=ApplicationType.DotBot,
            version=PROTOCOL_VERSION,
        )

    def update(self):
        """State space model update."""
        pos_x_old = self.pos_x
        pos_y_old = self.pos_y
        theta_old = self.theta

        if self.controller_mode == DotBotSimulatorMode.MANUAL:
            self.pos_x, self.pos_y, self.theta = diff_drive_bot(
                pos_x_old, pos_y_old, theta_old, self.v_right, self.v_left
            )
        elif self.controller_mode == DotBotSimulatorMode.AUTOMATIC:
            delta_x = self.pos_x - self.waypoints[self.waypoint_index][0]
            delta_y = self.pos_y - self.waypoints[self.waypoint_index][1]
            distance_to_target = sqrt(delta_x**2 + delta_y**2)

            # check if we are close enough to the "next" waypoint
            if distance_to_target < self.waypoint_threshold:
                self.logger.info("Waypoint reached", waypoint_index=self.waypoint_index)
                self.waypoint_index += 1
                # check if there are no more waypoints:
                if self.waypoint_index >= len(self.waypoints):
                    self.logger.info(
                        "Last waypoint reached", waypoint_index=self.waypoint_index
                    )
                    self.v_left = 0
                    self.v_right = 0
                    self.waypoint_index = 0
                    self.controller_mode = DotBotSimulatorMode.MANUAL
            else:
                robot_angle = self.theta
                angle_to_target = atan2(delta_y, delta_x)
                if robot_angle >= pi:
                    robot_angle = robot_angle - 2 * pi
                # if (angle_to_target < 0):
                #    angle_to_target = 2*pi + angle_to_target

                error_angle = ((angle_to_target - robot_angle + pi) % (2 * pi)) - pi
                self.logger.debug(
                    "Moving to waypoint",
                    robot_angle=robot_angle,
                    angle_to_target=angle_to_target,
                    error_angle=error_angle,
                )

                angular_speed = error_angle * 200
                self.v_left = 100 + angular_speed
                self.v_right = 100 - angular_speed

                if self.v_left > 100:
                    self.v_left = 100
                if self.v_right > 100:
                    self.v_right = 100
                if self.v_left < 0:
                    self.v_left = 0
                if self.v_right < 0:
                    self.v_right = 0

            self.pos_x, self.pos_y, self.theta = diff_drive_bot(
                self.pos_x, self.pos_y, self.theta, self.v_right, self.v_left
            )

        return self.encode_serial_output()

    def advertise(self):
        """Send an adertisement message to the gateway."""
        payload = ProtocolPayload(
            self.header,
            PayloadType.ADVERTISEMENT,
            Advertisement(),
        )
        return hdlc_encode(payload.to_bytes())

    def decode_serial_input(self, frame):
        """Decode the serial input received from the gateway."""
        payload = ProtocolPayload.from_bytes(hdlc_decode(frame))

        if self.address == hex(payload.header.destination)[2:]:
            if payload.payload_type == PayloadType.CMD_MOVE_RAW:
                self.controller_mode = DotBotSimulatorMode.MANUAL
                self.v_left = payload.values.left_y
                self.v_right = payload.values.right_y

                if self.v_left > 127:
                    self.v_left = self.v_left - 256
                if self.v_right > 127:
                    self.v_right = self.v_right - 256

            elif payload.payload_type == PayloadType.LH2_WAYPOINTS:
                self.controller_mode = DotBotSimulatorMode.MANUAL
                self.waypoint_threshold = payload.values.threshold * 1000
                self.waypoints = payload.values.waypoints
                if self.waypoints:
                    self.controller_mode = DotBotSimulatorMode.AUTOMATIC

    def encode_serial_output(self):
        """Encode the dotbot data to be sent to the gateway."""
        payload = ProtocolPayload(
            self.header,
            PayloadType.DOTBOT_SIMULATOR_DATA,
            DotBotSimulatorData(
                int(self.theta * 180 / pi + 90),
                int(self.pos_x),
                int(self.pos_y),
            ),
        )
        return hdlc_encode(payload.to_bytes())


class DotBotSimulatorSerialInterface(threading.Thread):
    """Bidirectional serial interface to control simulated robots"""

    def __init__(self, callback: Callable):
        self.dotbots = [
            DotBotSimulator("1234567890123456"),
            DotBotSimulator("4987654321098765"),
        ]

        self.callback = callback
        super().__init__(daemon=True)
        self.start()
        self.logger = LOGGER.bind(context=__name__)
        self.logger.info("DotBot Simulation Started")

    def run(self):
        """Listen continuously at each byte received on the fake serial interface."""
        advertising_intervals = [0] * len(self.dotbots)
        for dotbot in self.dotbots:
            for byte in dotbot.advertise():
                self.callback(byte.to_bytes(length=1, byteorder="little"))
        time.sleep(0.5)

        while 1:
            for idx, dotbot in enumerate(self.dotbots):
                for byte in dotbot.update():
                    self.callback(byte.to_bytes(length=1, byteorder="little"))
                    time.sleep(0.005)
                advertising_intervals[idx] += 1
                if advertising_intervals[idx] == 100:
                    for byte in dotbot.advertise():
                        self.callback(byte.to_bytes(length=1, byteorder="little"))
                        time.sleep(0.005)
                    advertising_intervals[idx] = 0
            time.sleep(0.02)

    def write(self, bytes_):
        """Write bytes on the fake serial."""
        for dotbot in self.dotbots:
            dotbot.decode_serial_input(bytes_)



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Dotbot controller serial interface."""

import sys
import threading
import time
from typing import Callable

import serial
from serial.tools import list_ports

from dotbot.logger import LOGGER

PAYLOAD_CHUNK_SIZE = 64
PAYLOAD_CHUNK_DELAY = 0.002  # 2 ms


def get_default_port():
    """Return default serial port."""
    ports = [port for port in list_ports.comports()]
    if sys.platform != "win32":
        ports = [port for port in ports if "J-Link" == port.product]
    if not ports:
        return "/dev/ttyACM0"
    # return first JLink port available
    return ports[0].device


class SerialInterfaceException(Exception):
    """Exception raised when serial port is disconnected."""


class SerialInterface(threading.Thread):
    """Bidirectional serial interface."""

    def __init__(self, port: str, baudrate: int, callback: Callable):
        self.callback = callback
        self.serial = serial.Serial(port, baudrate)
        super().__init__(daemon=True)
        self.start()
        self._logger = LOGGER.bind(context=__name__)
        self._logger.info("Serial port thread started")

    def run(self):
        """Listen continuously at each byte received on serial."""
        try:
            while 1:
                try:
                    byte = self.serial.read(1)
                except (TypeError, serial.serialutil.SerialException):
                    byte = None
                if byte is None:
                    self._logger.info("Serial port disconnected")
                    break
                self.callback(byte)
        except serial.serialutil.PortNotOpenError as exc:
            self._logger.error(f"{exc}")
            raise SerialInterfaceException(f"{exc}") from exc
        except serial.serialutil.SerialException as exc:
            self._logger.error(f"{exc}")
            raise SerialInterfaceException(f"{exc}") from exc

    def stop(self):
        self.serial.close()
        self.join()

    def write(self, bytes_):
        """Write bytes on serial."""
        # Send 64 bytes at a time
        pos = 0
        while (pos % PAYLOAD_CHUNK_SIZE) == 0 and pos < len(bytes_):
            self.serial.write(bytes_[pos : pos + PAYLOAD_CHUNK_SIZE])
            self.serial.flush()
            pos += PAYLOAD_CHUNK_SIZE
            time.sleep(PAYLOAD_CHUNK_DELAY)



import os

# Configuration file for the Sphinx documentation builder.

project = "PyDotBot"
copyright = "2023, Inria"
author = "Alexandre Abadie"

# -- General configuration ---------------------------------------------------

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.githubpages",
    "sphinx.ext.graphviz",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]

tls_verify = False
templates_path = ["_templates"]
exclude_patterns = ["_build"]
nitpick_ignore_regex = [
    ("py:class", r"abc.*"),
    ("py:class", r"enum.*"),
    ("py:class", r"numpy.*"),
    ("py:class", r"pydantic.*"),
    ("py:class", r"pynput.*"),
    ("py:class", r"threading.*"),
    ("py:class", r"starlette.*"),
    ("py:class", r"ConfigDict"),
    ("py:class", r"DotenvType"),
    ("py:class", r"FieldInfo"),
    ("py:class", r"Path"),
    ("py:class", r"SettingsConfigDict"),
    ("py:class", r"<MagicMock.*>"),
    ("py:class", r"ComputedFieldInfo"),
]

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_sourcelink_suffix = ""
html_static_path = ["_static"]

myst_enable_extensions = ["html_image"]

# Define the json_url for our version switcher.
json_url = "https://pydotbot.readthedocs.io/en/latest/_static/switcher.json"
rtd_version = os.environ.get("READTHEDOCS_VERSION")
rtd_version_type = os.environ.get("READTHEDOCS_VERSION_TYPE")
rtd_git_identifier = os.environ.get("READTHEDOCS_GIT_IDENTIFIER")
# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
# If it's "latest" → change to "dev" (that's what we want the switcher to call it)
if not rtd_version or rtd_version.isdigit() or rtd_version == "latest":
    rtd_version = "dev"
    json_url = "_static/switcher.json"
elif rtd_version == "stable":
    rtd_version = f"{rtd_git_identifier}"
elif rtd_version_type == "tag":
    rtd_version = f"{rtd_git_identifier}"

html_theme_options = {
    "external_links": [
        {
            "url": "https://github.com/DotBots/DotBot-firmware",
            "name": "DotBot firmware",
            "attributes": {
                "target": "_blank",
                "rel": "noopener me",
            },
        },
        {
            "url": "https://github.com/DotBots/DotBot-hardware",
            "name": "DotBot hardware",
            "attributes": {
                "target": "_blank",
                "rel": "noopener me",
            },
        },
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/DotBots/PyDotBot",
            "icon": "fa-brands fa-github",
        },
    ],
    "header_links_before_dropdown": 3,
    "logo": {
        "text": "PyDotBot",
    },
    "navbar_align": "left",
    "navbar_center": ["version-switcher", "navbar-nav"],
    "switcher": {
        "json_url": json_url,
        "version_match": rtd_version,
    },
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
}

# -- Options for linkcheck ---------------------------------------------

linkcheck_ignore = [r"http://localhost:\d+/"]

# -- Options for autosummary/autodoc output -----------------------------------
autosummary_generate = True
autodoc_typehints = "description"
autodoc_member_order = "groupwise"
